def autopage_fn(i, response, context):
    limit = context.params.get('limit')

    if (response.offset+1) < response.total_length:
        if limit is not None:
            if isinstance(context.params.get('offset'), int):
                context.params['offset'] += limit
            else:
                context.params['offset'] = limit

            if context.params['offset'] > response.total_length:
                return True

            return False

    return True


def make_response(results,
                  status_code=200,
                  metadata=None,
                  limit=None,
                  offset=None,
                  total=None,
                  parameters={}):
    payload = {
        'StatusCode': status_code,
    }

    if status_code < 400:
        payload['Status'] = 'success'
    else:
        payload['Status'] = 'error'

    if results is not None:
        if isinstance(results, list):
            payload['Results'] = results
        else:
            payload['Results'] = [results]

        payload['ResultCount'] = {}

        if total is not None:
            payload['ResultCount']['Total'] = total
        else:
            payload['ResultCount']['Total'] = len(payload['Results'])

        if len(payload['Results']) < payload['ResultCount']['Total']:
            payload['ResultCount']['Current'] = len(payload['Results'])

        if isinstance(limit, int):
            payload['ResultCount']['Limit'] = limit
            payload['ResultCount']['Offset'] = 0

        if isinstance(offset, int):
            payload['ResultCount']['Offset'] = offset

    if isinstance(metadata, dict):
        payload['Metadata'] = metadata

    payload['Parameters'] = parameters

    return payload



from __future__ import absolute_import
from unittest import TestCase
from .dicts import mutate_dict, \
                   camelize_dict, \
                   underscore_dict, \
                   compact


class UtilsDictTest(TestCase):
    def test_mutate_dict_identity(self):
        self.assertEqual(mutate_dict({
            'A': 1,
            u'b': True,
            False: 3.0,
            3.5: {
                7.0: [4, 5, 6],
            },
        }), {
            'A': 1,
            u'b': True,
            False: 3.0,
            3.5: {
                7.0: [4, 5, 6],
            },
        })

    def test_mutate_dict_keys_and_values(self):
        inval = {
            'aaa': 1,
            'bbb': 2,
            'ccc': 3.0,
            'ddd': [{
                'e': 4,
                'f': 5,
            }],
        }

        self.assertEqual(mutate_dict(inval, keyFn=lambda k: k.title()), {
            'Aaa': 1,
            'Bbb': 2,
            'Ccc': 3.0,
            'Ddd': [{
                'E': 4,
                'F': 5,
            }],
        })

        inval[1.0] = 9

        self.assertEqual(mutate_dict(inval,
                                     keyFn=lambda k: k.title(),
                                     keyTypes=(str, unicode)), {
            'Aaa': 1,
            'Bbb': 2,
            'Ccc': 3.0,
            'Ddd': [{
                'E': 4,
                'F': 5,
            }],
            1.0: 9,
        })

        inval['ddd'][0]['g'] = [6, 7, 8]

        self.assertEqual(mutate_dict(inval,
                                     valueFn=lambda v: (v*10),
                                     valueTypes=(int, long)), {
            'aaa': 10,
            'bbb': 20,
            'ccc': 3.0,
            'ddd': [{
                'e': 40,
                'f': 50,
                'g': [6, 7, 8],
            }],
            1.0: 90,
        })

        self.assertEqual(mutate_dict(inval,
                                     valueFn=lambda v: (v*10.0),
                                     valueTypes=(float,)), {
            'aaa': 1,
            'bbb': 2,
            'ccc': 30.0,
            'ddd': [{
                'e': 4,
                'f': 5,
                'g': [6, 7, 8],
            }],
            1.0: 9,
        })

    def test_camelize_dict(self):
        self.assertEqual(camelize_dict({
            'id':      1,
            'first_name': 'Fred',
            'last_name': 'Johnson',
            'list Of numbers': [1, 4, 7],
            'Things that-are': {
                'active': True,
                'to': [{
                    'be': 'done',
                }]
            },
            True: 'yep',
            5: '2+2',
        }, True), {
            'Id':      1,
            'FirstName': 'Fred',
            'LastName': 'Johnson',
            'ListOfNumbers': [1, 4, 7],
            'ThingsThatAre': {
                'Active': True,
                'To': [{
                    'Be': 'done',
                }]
            },
            True: 'yep',
            5: '2+2',
        })

    def test_camelize_dict_unicode_keys(self):
        self.assertEqual(camelize_dict({
            u'id':      1,
            u'first_name': 'Fred',
            u'last_name': 'Johnson',
            u'list Of numbers': [1, 4, 7],
            u'Things that-are': {
                u'active': True,
                u'to': [{
                    u'be': 'done',
                }]
            },
            True: 'yep',
            5: '2+2',
        }, True), {
            'Id':      1,
            'FirstName': 'Fred',
            'LastName': 'Johnson',
            'ListOfNumbers': [1, 4, 7],
            'ThingsThatAre': {
                'Active': True,
                'To': [{
                    'Be': 'done',
                }]
            },
            True: 'yep',
            5: '2+2',
        })

    def test_camelize_dict_list_of_dicts(self):
        self.assertEqual(camelize_dict([
            {'primary_key': 1},
            {'primary_key': 2},
            {'primary_key': 3},
        ]), [
            {'primaryKey': 1},
            {'primaryKey': 2},
            {'primaryKey': 3},
        ])

    def test_camelize_dict_passthrough(self):
        self.assertEqual(camelize_dict(1), 1)
        self.assertEqual(camelize_dict(True), True)
        self.assertEqual(camelize_dict(False), False)
        self.assertEqual(camelize_dict(3.14), 3.14)
        self.assertEqual(camelize_dict('lol'), 'lol')
        self.assertEqual(camelize_dict([]), [])

    def test_underscore_dict(self):
        self.assertEqual(underscore_dict({
            'Id':      1,
            'FirstName': 'Fred',
            'LastName': 'Johnson',
            'ListOfNumbers': [1, 4, 7],
            'ThingsThatAre': {
                'Active': True,
                'To': [{
                    'Be': 'done',
                }]
            },
            True: 'yep',
            5: '2+2',
        }), {
            'id':      1,
            'first_name': 'Fred',
            'last_name': 'Johnson',
            'list_of_numbers': [1, 4, 7],
            'things_that_are': {
                'active': True,
                'to': [{
                    'be': 'done',
                }]
            },
            True: 'yep',
            5: '2+2',
        })

    def test_underscore_dict_unicode_keys(self):
        self.assertEqual(underscore_dict({
            u'Id':      1,
            u'FirstName': 'Fred',
            u'LastName': 'Johnson',
            u'ListOfNumbers': [1, 4, 7],
            u'ThingsThatAre': {
                u'Active': True,
                u'To': [{
                    u'Be': 'done',
                }]
            },
            True: 'yep',
            5: '2+2',
        }), {
            u'id':      1,
            u'first_name': 'Fred',
            u'last_name': 'Johnson',
            u'list_of_numbers': [1, 4, 7],
            u'things_that_are': {
                u'active': True,
                u'to': [{
                    u'be': 'done',
                }]
            },
            True: 'yep',
            5: '2+2',
        })

    def test_underscore_dict_list_of_dicts(self):
        self.assertEqual(underscore_dict([
            {'primaryKey': 1},
            {'primaryKey': 2},
            {'primaryKey': 3},
        ]), [
            {'primary_key': 1},
            {'primary_key': 2},
            {'primary_key': 3},
        ])

    def test_underscore_dict_passthrough(self):
        self.assertEqual(underscore_dict(1), 1)
        self.assertEqual(underscore_dict(True), True)
        self.assertEqual(underscore_dict(False), False)
        self.assertEqual(underscore_dict(3.14), 3.14)
        self.assertEqual(underscore_dict('lol'), 'lol')
        self.assertEqual(underscore_dict([]), [])

    def test_compact_default(self):
        self.assertEqual(compact({
            'a': 1,
            'b': 2,
            'c': 3,
        }), {
            'a': 1,
            'b': 2,
            'c': 3,
        })

        self.assertEqual(compact({
            'a': 1,
            'b': None,
            'c': 3,
        }), {
            'a': 1,
            'c': 3,
        })

        self.assertEqual(compact({
            'a': None,
            'b': None,
            'c': None,
        }), {})

    def test_compact_custom(self):
        self.assertEqual(compact({
            'a': 1,
            'b': 2,
            'c': 3,
        }, lambda k, v: v > 0), {
            'a': 1,
            'b': 2,
            'c': 3,
        })

        self.assertEqual(compact({
            'a': 1,
            'b': None,
            'c': 3,
        }, lambda k, v: v > 1), {
            'c': 3,
        })

        self.assertEqual(compact({
            'a': 1,
            'b': 2,
            'c': 3,
        }, lambda k, v: v > 2), {
            'c': 3,
        })



"""
Functions for retrieving data from and manipulating dictionaries.
"""
from __future__ import absolute_import
from .strings import camelize, underscore

DATA_ELEMENT_SEPARATOR = '/'


class MissingKeyException(Exception):
    pass


def must_deep_get(data, path):
    """
    Retrieve a deeply-nested dict key, raise an exception if it does not exist.

    Args:
        path (str): A slash (/) separated string that is used to locate a given deeply
            nested dict key.

    Returns:
        The requested dict key value.

    Raises:
        :class:`~performline.utils.dicts.MissingKeyException`
    """
    if not isinstance(data, dict):
        raise Exception("Response data is unavailable")

    current = data
    parts = path.split(DATA_ELEMENT_SEPARATOR)

    for i in range(0, len(parts)):
        if isinstance(current, dict) and parts[i] in current:
            current = current[parts[i]]
        else:
            raise MissingKeyException("Cannot retrieve dict element '%s'" % path)

    return current


def deep_get(data, path, fallback=None):
    """
    Retrieve a deeply-nested dict key, return ``fallback`` if it does not exist.

    Args:
        path (str): A slash (/) separated string that is used to locate a given deeply
            nested dict key.

        fallback (object, optional): A value to return if the specified path does not exist.

    Returns:
        The requested dict key value, or ``fallback``.
    """
    try:
        return must_deep_get(data, path)
    except MissingKeyException:
        return fallback


def camelize_dict(inValue, upperFirst=False):
    """
    Recursively applies the :func:`~performline.utils.strings.camelize` function to all key names
    in the given dictionary.

    Args:

        upperFirst (bool, optional): Whether the first letter in the camelized output should
            be capitalized (i.e. "PascalCase"), or should be lower case (i.e. "camelCase").

    Returns:
        The `dict` with changes applied.
    """

    return mutate_dict(inValue,
                       keyFn=camelize,
                       keyTypes=(str, unicode),
                       upperFirst=upperFirst)


def underscore_dict(inValue, joiner='_'):
    """
    Recursively applies the :func:`~performline.utils.strings.underscore` function to all key names
    in the given dictionary.

    Args:

        joiner (str, optional): The string used to join individual words in key names.

    Returns:
        The `dict` with changes applied.
    """

    return mutate_dict(inValue,
                       keyFn=underscore,
                       keyTypes=(str, unicode),
                       joiner=joiner)


def mutate_dict(inValue,
                keyFn=lambda k: k,
                valueFn=lambda v: v,
                keyTypes=None,
                valueTypes=None,
                **kwargs):
    """
    Takes an input dict or list-of-dicts and applies ``keyfn`` function to all of the keys in
    both the top-level and any nested dicts or lists, and ``valuefn`` to all

    If the input value is not of type `dict` or `list`, the value will be returned as-is.

    Args:
        inValue (any): The dict to mutate.

        keyFn (lambda): The function to apply to keys.

        valueFn (lambda): The function to apply to values.

        keyTypes (tuple, optional): If set, only keys of these types will be mutated
            with ``keyFn``.

        valueTypes (tuple, optional): If set, only values of these types will be mutated
            with ``valueFn``.

    Returns:
        A recursively mutated dict, list of dicts, or the value as-is (described above).
    """

    # this is here as a way of making sure that the various places where recursion is done always
    # performs the same call, preserving all arguments except for value (which is what changes
    # between nested calls).
    def recurse(value):
        return mutate_dict(value,
                           keyFn=keyFn,
                           valueFn=valueFn,
                           keyTypes=keyTypes,
                           valueTypes=valueTypes,
                           **kwargs)

    # handle dicts
    if isinstance(inValue, dict):
        # create the output dict
        outputDict = dict()

        # for each dict item...
        for k, v in inValue.items():
            # apply the keyFn to some or all of the keys we encounter
            if keyTypes is None or (isinstance(keyTypes, tuple) and isinstance(k, keyTypes)):
                # prepare the new key
                k = keyFn(k, **kwargs)

            # apply the valueFn to some or all of the values we encounter
            if valueTypes is None or (isinstance(valueTypes, tuple) and isinstance(v, valueTypes)):
                v = valueFn(v)

            # recurse depending on the value's type
            #
            if isinstance(v, dict):
                # recursively call mutate_dict() for nested dicts
                outputDict[k] = recurse(v)
            elif isinstance(v, list):
                # recursively call mutate_dict() for each element in a list
                outputDict[k] = [recurse(i) for i in v]
            else:
                # set the value straight up
                outputDict[k] = v

        # return the now-populated output dict
        return outputDict

    # handle lists-of-dicts
    elif isinstance(inValue, list) and len(inValue) > 0:
        return [recurse(i) for i in inValue]

    else:
        # passthrough non-dict value as-is
        return inValue


def compact(inDict, keep_if=lambda k, v: v is not None):
    """
    Takes a dictionary and returns a copy with elements matching a given lambda removed. The
    default behavior will remove any values that are `None`.

    Args:
        inDict (dict): The dictionary to operate on.

        keep_if (lambda(k,v), optional): A function or lambda that will be called for each
            (key, value) pair.  If the function returns truthy, the element will be left alone,
            otherwise it will be removed.
    """
    if isinstance(inDict, dict):
        return {
            k: v for k, v in inDict.items() if keep_if(k, v)
        }

    raise ValueError("Expected: dict, got: {0}".format(type(inDict)))



from __future__ import absolute_import
from unittest import TestCase
from .strings import camelize, underscore


class UtilsStringTest(TestCase):
    def test_strings_camelize(self):
        """
        Test if camelize actually camelizes various strings as expected
        """

        self.assertEqual(camelize('test_value'),       'testValue')
        self.assertEqual(camelize('test-Value'),       'testValue')
        self.assertEqual(camelize('test value'),       'testValue')
        self.assertEqual(camelize('TestValue'),        'testValue')
        self.assertEqual(camelize('testValue'),        'testValue')
        self.assertEqual(camelize('TeSt VaLue'),       'teStVaLue')

        self.assertEqual(camelize('test_value', True), 'TestValue')
        self.assertEqual(camelize('test-Value', True), 'TestValue')
        self.assertEqual(camelize('test value', True), 'TestValue')
        self.assertEqual(camelize('TestValue',  True), 'TestValue')
        self.assertEqual(camelize('testValue',  True), 'TestValue')
        self.assertEqual(camelize('TeSt VaLue', True), 'TeStVaLue')

    def test_strings_underscore(self):
        """
        Test if underscore actually underscores various strings as expected
        """

        self.assertEqual(underscore('test_value'),                   'test_value')
        self.assertEqual(underscore('test-Value'),                   'test_value')
        self.assertEqual(underscore('test value'),                   'test_value')
        self.assertEqual(underscore('TestValue'),                    'test_value')
        self.assertEqual(underscore('testValue'),                    'test_value')
        self.assertEqual(underscore('TeSt VaLue'),                   'te_st_va_lue')

        self.assertEqual(underscore('test_value', joiner='-'),       'test-value')
        self.assertEqual(underscore('test-Value', joiner='-'),       'test-value')
        self.assertEqual(underscore('test value', joiner='-'),       'test-value')
        self.assertEqual(underscore('TestValue', joiner='-'),        'test-value')
        self.assertEqual(underscore('testValue', joiner='-'),        'test-value')
        self.assertEqual(underscore('TeSt VaLue', joiner='-'),       'te-st-va-lue')



from __future__ import absolute_import
import json
from ...utils.dicts import deep_get, camelize, camelize_dict, underscore_dict
from ...utils.dicts import compact as utils_compact
from .exceptions import UnsupportedOperation


class RestModel(object):
    """
    Represents individual objects that can be read, created, and/or update via a given client.

    Args:
        rest_root (str, abstract): The path that will be called when operating on this model
            object.  All interactions with objects of this type will have this path prepended to
            them when making HTTP calls.  For operations on specific instances, this value will be
            combined with ``primary_key`` to form the final path.  For example:

            If ``rest_root = '/example/'`` and ``primary_key = 'id'``, the pattern for retrieving a
            single object will be "/example/{id}/".  This path will will then have the `format()`
            method called on it whenever it is used with the current value of the ``data`` property
            passed in as `**kwargs`.  This allows data values to be interpolated into the path at
            call time.

            This value must be set (explictly during instantiation or implicitly in a subclass)
            before the class can be used.

        rest_read_method (str, optional): The HTTP method that will be used for retrieving the
            object this instance represents.

        rest_create_method (str, optional): The HTTP method that will be used for creating the
            object this instance represents.

        rest_update_method (str, optional): The HTTP method that will be used for updating the
            object this instance represents.
    """

    primary_key = 'id'
    rest_read_method = 'get'
    rest_create_method = 'post'
    rest_update_method = 'put'

    def __init__(self, client, data=None, metadata={}, rest_root=None, primary_key=None):
        """
        Provides a simplified interface for working with standard REST API responses.

        Args:
            client (:class:`~performline.clients.rest.StandardRestClient`): The REST client that
                was used to retrieve the data or one that will be used to persist it.

            data (dict): The dictionary that represents this model.

            rest_root (str, optional): If specified, this will set or override the base path this
                instance uses to retrieve or persist data.

            primary_key (str, optional): If specified, this field name will be used for all
                operations that occur on a specific instance of this object.
        """

        if rest_root is not None:
            self.rest_root = rest_root

        if primary_key is not None:
            self.primary_key = primary_key

        if not hasattr(self, 'rest_root') or self.rest_root is None:
            raise AttributeError("Cannot create RestModel instance without a "
                                 "valid 'rest_root' attribute")

        if not hasattr(self, 'primary_key') or self.primary_key is None:
            raise AttributeError("Cannot create RestModel instance without a "
                                 "valid 'primary_key' attribute")

        self.__metadata = metadata
        self._client = client

        if data is not None:
            self.set_data(data)

    @classmethod
    def get(cls, client, pk):
        return cls(client, {
            cls.primary_key: pk,
        }).retrieve()

    @classmethod
    def all(cls, client, autoload=True, **kwargs):
        items = []

        for response in client.get_until(cls.rest_root, **kwargs):
            # for each result
            for item in response.results():
                # format the list result item the same way formatted_path would
                item = underscore_dict(item)

                # get the value of the primary key field
                pk = item.get(cls.primary_key)

                if pk is not None:
                    instance = cls(client, {
                        cls.primary_key: pk,
                    })

                    if autoload:
                        instance.retrieve()

                    items.append(instance)

        return items

    @property
    def pk(self):
        return self._data.get(self.primary_key)

    @property
    def _data(self):
        """
        Retrieves the underlying data that this model represents.

        Returns:
            dict
        """
        if self.__data is None:
            return {}
        return self.__data

    @property
    def _metadata(self):
        return self.__metadata

    @property
    def instance_path(self):
        return self.rest_root.rstrip('/') + '/{' + self.primary_key.strip('/') + '}/'

    def formatted_path(self, additional_args={}):
        """
        Returns the ``instance_path`` with data (and, if specified, additional/override arguments)
        interpolated in.  This method uses the standard Python ``format()`` function.

        Returns:
            str
        """

        kwa = {}

        for k, v in underscore_dict(self._data).items():
            kwa[k] = getattr(self, k)

        kwa.update(additional_args)
        return self.instance_path.format(**kwa)

    def set_data(self, data):
        """
        Updates the model's underlying data, processing it with :func:`prepare_data_from_retrieval`
        first.
        """
        self.__data = self.prepare_data_from_retrieval(data)

    def set_metadata(self, metadata):
        """
        Updates the model's metadata, processing it with :func:`prepare_metadata_from_retrieval`.
        first.
        """
        self.__metadata = self.prepare_metadata_from_retrieval(metadata)

    def prepare_data_from_retrieval(self, data):
        """
        A method used to prepare data as retrieved from the client in response to a retrieve (GET)
        request.  This default implementation processes the data through
        :func:`~performline.utils.dicts.camelize_dict`, but can be overridden in a subclass to
        perform custom post-processing.

        Returns:
            dict of post-processed data.
        """
        return camelize_dict(data, True)

    def prepare_metadata_from_retrieval(self, metadata):
        """
        A method used to prepare metadata as retrieved from the client in response to a
        retrieve (GET) request.  This default implementation passes the metadata to
        :func:`prepare_data_from_retrieval`, but can be overridden in a subclass to perform custom
        post-processing.

        Returns:
            dict of post-processed metadata.
        """

        return self.prepare_data_from_retrieval(metadata)

    def prepare_data_for_update(self, data):
        """
        A method used to prepare the model's data for submission to the service (POST/PUT). This
        default implementation is a no-op and returns the given ``data`` as-is, but can be
        overridden in a subclass to perform custom pre-processing.

        Args:
            data (dict): The model's data about to be sent to the upstream service.

        Returns:
            dict of pre-processed data.
        """
        return data

    @property
    def client(self):
        """
        Retrieves the client object that is used for interacting with services.

        Returns:
            :class:`~performline.clients.rest.StandardRestClient`
        """
        return self._client

    def retrieve(self):
        """
        Refreshes this model's internal representation with the server.
        """

        if self.rest_read_method is None:
            raise UnsupportedOperation("Retrieval is not supported on {0}".format(self.__class__))

        response = self.client.request(self.rest_read_method, self.formatted_path())
        result = response.results(0)

        self.set_data(result)
        self.set_metadata(response.metadata)

        return self

    def save(self, refresh=True, compact=False):
        """
        Persists this model's current data to the server.

        Args:
            refresh (bool): Whether to automatically refresh the model with the data that was
                just persisted.

            compact (bool, lambda, optional): Whether the data should be compacted before being
                submitted to the server.  If this is True, _None_ values will be removed.  If this
                is a lambda matching the signature ``f(key,value) -> bool``, that will be the
                function that determines which keys are kept (return True) and which are rejected
                (return False).

        Raises:
            AttributeError if the given ``key_attr`` is not a valid attribute on this instance.
        """

        if self.rest_create_method is None and self.rest_update_method is None:
            raise UnsupportedOperation("Persistence is not supported on {0}"
                                       .format(self.__class__))

        data = self._data

        # compacts the data (either with the default method or with a user-specified one)
        if compact:
            if isinstance(compact, type(lambda: 0)):
                data = utils_compact(data, compact)
            else:
                data = utils_compact(data)

        data = self.prepare_data_for_update(data)

        # performs the request
        self.client.request(self.rest_update_method,
                            self.formatted_path(),
                            data)

        # optionally refresh (read back) the data we just saved
        if refresh:
            return self.retrieve()

        return True

    def __getattr__(self, name):
        """
        Retrieves the named attribute as if it were a property if it exists in this model's
        response payload.  Otherwise, the attribute behaves normally.  Attributes that start
        with an underscore (_) will be passed along to the normal attribute handler as usual.

        Returns:
            any
        """
        if not name.startswith('_'):
            candidate_key = camelize(name, True)
            value = deep_get(self._data, name, deep_get(self._data, candidate_key))

            if value is not None:
                return value

    def __setattr__(self, name, value):
        key = camelize(name, True)

        if isinstance(self._data, dict) and key in self._data:
            print('Setting %s' % key)
            self._data[key] = value

        object.__setattr__(self, name, value)

    def __iter__(self):
        for key, value in self._data.items():
            yield (key, value)

    def __dict__(self):
        return self._data

    def __repr__(self):
        return json.dumps(self._data, indent=4)
    __str__ = __repr__



"""
PerformLine Standard Library
"""
# makes this a "namespace package"
# this must be present in the __init__.py of any "performline" package
# that overlaps with this one.
from pkgutil import extend_path
__path__ = extend_path(__path__, __name__)  # noqa

version = '0.6.10'
"""str: The current version of the library implementation."""



"""
Python type system extensions and utilities.
"""
import six


PRIMITIVE_TYPES = (bool, int, float, complex, str, list, tuple, dict)

if six.PY2:
    PRIMITIVE_TYPES += (long, unicode)


def isprimitive(value):
    """
    Returns whether a given value is a primitive type or not.

    Args:
        value (any): The value to test.

    Returns:
        bool
    """

    if value is None or isinstance(value, PRIMITIVE_TYPES):
        return True

    return False



# Copyright (c) 2016, PerformLine, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the company nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL PERFORMLINE, INC. BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from __future__ import absolute_import
import click
from ....cliutils import out


@click.group(help='Call transcripts registered for processing')
def calls():
    pass


@calls.command(help='List all calls')
@click.pass_obj
def list(state):
    out(state, state.client.calls())


@calls.command(help='Show details about a single call')
@click.argument('id',
                type=int)
@click.pass_obj
def show(state, id):
    out(state, state.client.calls(id))



# Copyright (c) 2016, PerformLine, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the company nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL PERFORMLINE, INC. BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from __future__ import absolute_import
from .embedded.stdlib.clients.rest import StandardRestClient
from .embedded.stdlib.clients.rest.models import RestModel
from .embedded.stdlib.clients.rest.responses import SuccessResponse
from .products.common.api import CommonClientMethods
from .products.web.api import WebClientMethods
from .products.callcenter.api import CallCenterClientMethods
from .products.chatscout.api import ChatScoutClientMethods


ALLOWED_METHODS = ['get', 'post', 'put', 'delete', 'options', 'head']
RX_TRIM_SLASH = r'(?:^/*|/*$)'


class Client(
    ChatScoutClientMethods,
    CallCenterClientMethods,
    WebClientMethods,
    CommonClientMethods,
    StandardRestClient
):
    """PerformLine API Client"""

    url = 'https://api.performline.com'
    prefix = '/'

    def __init__(self, token, *args, **kwargs):
        self.headers['Authorization'] = 'Token {0}'.format(token)
        super(Client, self).__init__(*args, **kwargs)

    def wrap_response(self, response, model, flat=False):
        """
        Takes a SuccessResponse and returns the results as instances of the
        given model.

        Args:
            response (SuccessResponse): A SuccessResponse object containing the
                data to be wrapped.

            model (RestModel): A subclass of RestModel that should be used when creating
                instances for each result in the response.

            flat (bool, optional): Whether single-element lists should return
                just that element instead of the list itself.

        Returns:
            If ``response`` contains zero results, returns `None`.

            If ``response`` has one element and ``flat`` equals `True`, return the first
            element as an instance of type ``model``.

            In all other cases, return a list of instances of type ``model``.
        """

        if isinstance(response, SuccessResponse) and response.length > 0:
            if issubclass(model, RestModel):
                models = [model(self, i) for i in response.results()]

                if flat and len(models) == 1:
                    return models[0]
                else:
                    return models

        return None



from __future__ import absolute_import
from datetime import datetime
from .responses import Response


class TooManyIterations(Exception):
    pass


class UnsupportedOperation(Exception):
    pass


class ErrorResponse(Response, Exception):
    """
    A subclass of :class:`~performline.clients.rest.responses.Response` that provides accessors for
    fields present in erroneous REST responses.  It is also a subclass of :class:`Exception` and
    can be raised as such.
    """
    data = {}

    def __init__(self, response=None, data=None, exception=None, message=None):
        if data is not None:
            self.data = data

        self.message = message or 'Unknown Error'

        if response is not None:
            self.response = response
            self.message = self.deep_get('ErrorMessage')
            self.traceback = self.deep_get('Traceback', [])
        elif message is None and exception is not None:
            self.message = exception.message

        super(Exception, self).__init__(self.message)


class AuthenticationFailed(ErrorResponse):
    """
    An error condition indicating that the provided API key is not authorized to access
    a given resource.
    """
    code = 403
    data = {
        'Status': 'error',
        'StatusCode': 403,
    }


class NotFound(ErrorResponse):
    """
    An error condition indicating that the requested resource does not exist.
    """
    code = 404
    data = {
        'Status': 'error',
        'StatusCode': 404,
    }


class ServiceUnavailable(ErrorResponse):
    """
    An error condition indicating that the API service is temporarily unavailable.
    """
    code = 503
    data = {
        'Status': 'error',
        'StatusCode': 503,
    }


class UninitializedClient(Exception):
    """
    Indicates that a client instance is required to perform an operation, but one was not provided.
    """
    pass


class DeprecatedEndpoint(ErrorResponse):
    """
    Indicates that the requested endpoint is deprecated and should not be used.
    """
    code = 410
    data = {
        'Status': 'error',
        'StatusCode': 410,
    }

    def __init__(self, message, cutoff=None):
        self._cutoff = cutoff
        super(Exception, self).__init__(message)

    @property
    def cutoff(self):
        return self._cutoff

    def is_after_cutoff(self):
        """
        Returns whether the current UTC time is after the cutoff time (if available).

        Returns:
            bool
        """
        if self.cutoff:
            if datetime.utcnow() > self.cutoff:
                return True

        return False


class BadGateway(ErrorResponse):
    """
    Indicates an error message from an upstream service.
    """
    code = 502
    data = {
        'Status': 'error',
        'StatusCode': 502,
    }



from __future__ import absolute_import
from unittest import TestCase
from .responses import Response, SuccessResponse
from ...utils.dicts import MissingKeyException
from .utils import make_response


class RestResponseTest(TestCase):
    def test_response_data_access(self):
        response = Response(None, make_response([{
                'Id': 1,
                'Name': 'First',
                'Nested': {
                    'Here': 123,
                },
            }, {
                'Id': 2,
                'Name': 'Second',
            }],
            metadata={
                'this': {
                    'is': True,
                },
            },
            limit=10))

        self.assertIsInstance(response.payload, dict)
        self.assertEqual(response.response_status, 'success')
        self.assertEqual(response.status_code, 200)
        self.assertFalse(response.haserror())
        self.assertEqual(response.deep_get('ResultCount/Total'), 2)
        self.assertEqual(response.deep_get('ResultCount/Limit'), 10)
        self.assertEqual(response.deep_get('ResultCount/Offset'), 0)

        with self.assertRaises(MissingKeyException):
            response.must_deep_get('Nothing')

        with self.assertRaises(MissingKeyException):
            response.must_deep_get('Nothing/Offset')

        self.assertIsInstance(response.metadata, dict)
        self.assertEqual(response.metadata_get('this/is'), True)
        self.assertEqual(response.metadata_get('this/is/not'), None)
        self.assertEqual(response.metadata_get('this'), {
            'is': True,
        })

    def test_success_response_data_access(self):
        response = SuccessResponse(None, make_response([{
            'Id': 1,
            'Name': 'First',
            'Nested': {
                'Here': 123,
            },
        }, {
            'Id': 2,
            'Name': 'Second',
        }], metadata={
            'this': {
                'is': True,
            },
        }, total=25, limit=2, offset=4))

        self.assertEqual(response.total_length, 25)
        self.assertEqual(response.length, 2)
        self.assertEqual(response.limit, 2)
        self.assertEqual(response.offset, 4)
        self.assertEqual(response.total_pages, 13)
        self.assertEqual(response.current_page, 3)



from __future__ import absolute_import
from unittest import TestCase
from .types import isprimitive
import six


class UtilsTypesTest(TestCase):
    def test_isprimitive(self):
        self.assertTrue(isprimitive(True))
        self.assertTrue(isprimitive(False))
        self.assertTrue(isprimitive(int(1)))
        self.assertTrue(isprimitive(float(1.0)))
        self.assertTrue(isprimitive(5j))
        self.assertTrue(isprimitive('1'))
        self.assertTrue(isprimitive([1]))
        self.assertTrue(isprimitive((1,)))
        self.assertTrue(isprimitive({1: True}))

        if six.PY2:
            self.assertTrue(isprimitive(long(1)))
            self.assertTrue(isprimitive(u'1'))

        self.assertFalse(isprimitive(Exception("1")))
        self.assertFalse(isprimitive(self))
        self.assertFalse(isprimitive(TestCase))



# Copyright (c) 2016, PerformLine, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the company nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL PERFORMLINE, INC. BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from __future__ import absolute_import
from .models import Chat


class ChatScoutClientMethods(object):
    """Methods for retrieving data from the ChatScout product"""

    def chats(self, id=None, **kwargs):
        """
        Retrieve one or more chats associated with an account.

        Args:
            id (int, optional): If specified, retrieve a single chat by
                the given ID.  Otherwise, return all chats.

        Returns:
            An instance of :class:`~performline.products.chatscout.models.Chat`
            if ``id`` is not `None` representing the chat with that ID.
            Otherwise, retrieve a list of
            :class:`~performline.products.chatscout.models.Chat` instances of all
            chats associated with the account.

        Raises:
            See :func:`~performline.client.Client.request`
        """
        if id is None:
            return Chat.all(self, **kwargs)
        else:
            return Chat.get(self, id)



# Copyright (c) 2016, PerformLine, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the company nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL PERFORMLINE, INC. BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from __future__ import absolute_import
import click
import sys
import json
import yaml


def normalize(data):
    if hasattr(data, '__dict__'):
        data = data.__dict__()
    elif isinstance(data, list):
        if len(data) > 0:
            data = [normalize(i) for i in data]

    return data


def errout(message, exitcode=1):
    sys.stderr.write('%s\n' % message)
    exit(exitcode)


def out(state, data=None):
    rv = ''
    data = normalize(data)

    if data is None:
        return

    if isinstance(data, (list, dict)) and len(data) == 0:
        return

    if state.fmt == 'json':
        rv = json.dumps(data, indent=4)
    elif state.fmt == 'yaml':
        rv = yaml.safe_dump(data, default_flow_style=False)
    else:
        rv = str(data)

    if len(rv) > 0:
        click.echo(rv)



from __future__ import absolute_import
from unittest import TestCase
import json
import httpretty
from . import StandardRestClient
from .utils import make_response


class RestRequestTest(TestCase):
    def setUp(self):
        self.client = StandardRestClient(url='http://test-rest-service')

    def test_request_until(self):
        httpretty.enable()
        httpretty.register_uri(httpretty.GET,
                               ("http://test-rest-service/request_until_test/"
                                "?limit=1&offset=0"),
                               content_type='application/json',
                               responses=[
                                   httpretty.Response(body=json.dumps(make_response({
                                       'Id': 1,
                                   }, total=3, limit=1))),
                                   httpretty.Response(body=json.dumps(make_response({
                                       'Id': 2,
                                   }, total=3, limit=1, offset=1))),
                                   httpretty.Response(body=json.dumps(make_response({
                                       'Id': 3,
                                   }, total=3, limit=1, offset=2))),
                               ])

        responses = self.client.get_until('/request_until_test/', params={
            'limit': 1,
        })

        httpretty.disable()

        self.assertEqual(len(responses), 3)

        self.assertEqual(responses[0].results(0)['Id'], 1)
        self.assertEqual(responses[0].limit, 1)
        self.assertEqual(responses[0].offset, 0)

        self.assertEqual(responses[1].results(0)['Id'], 2)
        self.assertEqual(responses[1].limit, 1)
        self.assertEqual(responses[1].offset, 1)

        self.assertEqual(responses[2].results(0)['Id'], 3)
        self.assertEqual(responses[2].limit, 1)
        self.assertEqual(responses[2].offset, 2)



from __future__ import absolute_import
import logging
import requests
import requests.exceptions
import time
from datetime import datetime
import json
import yaml
from .responses import SuccessResponse
from .exceptions import ErrorResponse, \
                        AuthenticationFailed, \
                        NotFound, \
                        ServiceUnavailable, \
                        DeprecatedEndpoint, \
                        BadGateway, \
                        TooManyIterations
from .utils import autopage_fn


ALLOWED_METHODS = ['get', 'post', 'put', 'delete', 'options', 'head']
RX_TRIM_SLASH = r'(?:^/*|/*$)'


class RequestContext(object):
    def __init__(self, client, method, path, data=None, params={}, headers={}):
        self.client = client
        self.method = method
        self.path = path
        self.data = data
        self.params = params
        self.headers = headers


class StandardRestClient(object):
    """
    A standardized base class that can be used directly or as a subclass for consuming RESTful
    HTTP services at PerformLine.

    Args:
        url (str): The base URL that will apply to all instances of this class (can be overridden
            in subclasses)

        prefix (str): A string to prefix all paths given to the :func:`request()` method.  This is
            useful for endpoints that are always rooted at a particular path (e.g.: "/api").

        loglevel (str): A valid `logger` log level that is used for reporting details in the
            underlying HTTP library.

        content_type (str): The default MIME type that is presented to and expected from
            server responses.  This is a shorthand that populates the 'Content-Type' header by
            default.

        headers (dict): A dictionary of HTTP headers that should be included in every request. For
            example, an 'Authorization' header.

        params (dict): A dictionary of query string parameters that should be included in every
            request.  For example, `{'format': 'json'}` (which would include '?format=json' in the
            request URLs).
    """

    url = None
    prefix = '/'
    content_type = 'application/json'
    headers = {}
    params = {}
    request_options = {}

    def __init__(self,
                 url=None,
                 prefix=None,
                 loglevel='WARNING',
                 content_type=None,
                 params=None,
                 headers=None,
                 request_options={}):
        # check and set the various kwargs
        for p in [
            'url',
            'prefix',
            'content_type',
            'params',
            'headers',
            'request_options',
        ]:
            value = locals().get(p)
            if value is not None:
                setattr(self, p, value)

        # pass along loglevel to logging module
        if isinstance(loglevel, str):
            loglevel = loglevel.upper()

        lvl = logging.getLevelName(loglevel)

        if isinstance(lvl, int):
            self.loglevel = lvl
        else:
            self.loglevel = logging.INFO

        logging.basicConfig(level=self.loglevel)

        # disables the warnings requests emits, which ARE for our own good, but if we make the
        # decision to do something stupid, we'll own that and don't need to pollute the logs.
        requests.packages.urllib3.disable_warnings()

    def request(self, method, path, data=None, params={}, headers={}, encoder='json', **kwargs):
        """
        Perform a generic HTTP request against this instance's configured host.

        Args:
            method (str): The HTTP method to use.

            path (str): The path component of the URL for the request.

            data (any): Data to be included in the request body.

            params (dict): Query string parameters to included in the request URL.

            headers (dict): HTTP headers to include in the request.

            encoder (str, optional): If ``data`` is not None, which encoder should be used to
                pre-process the data.  A method on this client instance matching the
                format ``encode_<this value>(self,data,opts)`` will be called (see
                :func:`encode_json`, :func:`encode_yaml`). This allows for the easy implementation
                of custom encoders by defining (or overriding) methods matching this signature.
                The ``opts`` parameter will be a `dict` with keys matching some of the other
                parameters of this method (i.e.: `method`, `path`, `params`, and `headers`).  If
                this is `None`, encoding will be skipped.

            **kwargs: Additional arguments to pass directly to the underlying requests call.

        Returns:
            :class:`~performline.clients.rest.responses.SuccessResponse` if the HTTP response code
            is strictly less than 400.

        Raises:
            :class:`~performline.clients.rest.exceptions.ErrorResponse`
            :class:`~performline.clients.rest.exceptions.AuthenticationFailed`
            :class:`~performline.clients.rest.exceptions.NotFound`
            :class:`~performline.clients.rest.exceptions.ServiceUnavailable`
            :class:`~performline.clients.rest.exceptions.DeprecatedEndpoint`
        """

        if not method.lower() in ALLOWED_METHODS:
            raise Exception('Invalid request method {0}'.format(method))

        # normalize the given path
        path = path.strip('/')

        # set content type
        if self.content_type is not None:
            headers['Content-Type'] = self.content_type

        # merge in default headers and params
        params.update(self.params)
        headers.update(self.headers)

        # handle data encoding (if there is data to encode and we were asked to)
        if data is not None and encoder is not None:
            try:
                encodeFn = getattr(self, 'encode_%s' % encoder)
                data = encodeFn(data, {
                    'method': method,
                    'path': path,
                    'params': params,
                    'headers': headers,
                })
            except AttributeError:
                raise AttributeError('Cannot encode data: encoder "{0}" not implemented'
                                     .format(encoder))

        # merge client-global options into the kwargs
        if isinstance(self.request_options, dict):
            kwargs.update(self.request_options)

        # perform the request
        try:
            response = getattr(requests, method.lower())(
                '{0}/{1}{2}/'.format(
                    self.url.strip('/'),
                    self.prefix.lstrip('/'),
                    path),
                data=data,
                params=params,
                headers=headers,
                **kwargs)
        except requests.exceptions.SSLError:
            raise
        except requests.exceptions.ConnectionError as e:
            raise BadGateway(message="Failed to connect to '{0}'"
                             .format(e.request.url), exception=e)

        # check for endpoint deprecation, and raise an error if the endpoint cutoff date is
        # in the past.
        try:
            _check_response_for_deprecation(path, response)
        except DeprecatedEndpoint as e:
            logging.warning((
                'Upgrade the performline Python module to the latest version to '
                'stop seeing this message.'
            ))

            if e.is_after_cutoff():
                raise

        if response.status_code < 400:
            return SuccessResponse(response, response.json())
        else:
            try:
                body = response.json()
            except ValueError:
                body = {
                    'Status': 'error',
                    'StatusCode': response.status_code,
                    'ErrorMessage': response.text,
                }

            if response.status_code == 403:
                raise AuthenticationFailed(response, body)
            elif response.status_code == 404:
                raise NotFound(response, body)
            elif response.status_code >= 500:
                raise ServiceUnavailable(response, body)
            else:
                raise ErrorResponse(response, body)

    def get(self, *args, **kwargs):
        """
        Perform a GET request.  See: :func:`request`.
        """
        return self.request('get', *args, **kwargs)

    def post(self, *args, **kwargs):
        """
        Perform a POST request.  See: :func:`request`.
        """
        return self.request('post', *args, **kwargs)

    def put(self, *args, **kwargs):
        """
        Perform a PUT request.  See: :func:`request`.
        """
        return self.request('put', *args, **kwargs)

    def delete(self, *args, **kwargs):
        """
        Perform a DELETE request.  See: :func:`request`.
        """
        return self.request('delete', *args, **kwargs)

    def options(self, *args, **kwargs):
        """
        Perform an OPTIONS request.  See: :func:`request`.
        """
        return self.request('options', *args, **kwargs)

    def head(self, *args, **kwargs):
        """
        Perform a HEAD request.  See: :func:`request`.
        """
        return self.request('head', *args, **kwargs)

    def request_until(self, method, path, data=None, params={}, headers={}, encoder='json',
                      testfn=autopage_fn, request_delay_ms=0, max_iterations=25, **kwargs):
        """
        Perform an HTTP request repeatedly until a given function returns true.

        Args:
            See: :func:`request`

            testfn (func/lambda): A function that will be called with the signature
                ``fn(i, response, context)``; where:
                - ``i`` is the current iteration number (starting from 0)
                - ``response`` is the response for the HTTP request.
                - ``context`` (:class:`RequestContext`) is an editable request context that can
                    be used for modifying the request details for subsequent calls.

                The function should return a truthy value when request processing should stop.

            request_delay_ms (int): The number of milliseconds to wait between successive requests
                if ``testfn`` has not evaluated to True.

            max_iterations (int): If set to a value, this is the maximum number of iterations that
                will occur before returning regardless of the output of ``testfn``.
        """

        responses = []
        count = 0
        context = RequestContext(self, method, path, data, params, headers)

        while True:
            response = self.request(context.method,
                                    context.path,
                                    context.data,
                                    context.params,
                                    context.headers, encoder, **kwargs)

            responses.append(response)

            # break if we've exceeded max_iterations
            if max_iterations is not None:
                if count >= max_iterations:
                    raise TooManyIterations(("Request {0} {1} has exceeded the maximum iteration "
                                             "count of {2}").format(method.upper(),
                                                                    path,
                                                                    max_iterations))

            # only loop if we were given a callable testfn
            if callable(testfn):
                # call testfn to see if we should stop now
                if testfn(count, response, context):
                    break
            else:
                break

            count += 1

            # sleep if requested
            if request_delay_ms > 0:
                time.sleep(float(request_delay_ms) / 1000.0)

        return responses

    def get_until(self, *args, **kwargs):
        """
        Perform a GET request repeatedly.  See: :func:`request_until`.
        """
        return self.request_until('get', *args, **kwargs)

    def post_until(self, *args, **kwargs):
        """
        Perform a POST request repeatedly.  See: :func:`request_until`.
        """
        return self.request_until('post', *args, **kwargs)

    def put_until(self, *args, **kwargs):
        """
        Perform a PUT request repeatedly.  See: :func:`request_until`.
        """
        return self.request_until('put', *args, **kwargs)

    def delete_until(self, *args, **kwargs):
        """
        Perform a DELETE request repeatedly.  See: :func:`request_until`.
        """
        return self.request_until('delete', *args, **kwargs)

    def options_until(self, *args, **kwargs):
        """
        Perform a OPTIONS request repeatedly.  See: :func:`request_until`.
        """
        return self.request_until('options', *args, **kwargs)

    def head_until(self, *args, **kwargs):
        """
        Perform a HEAD request repeatedly.  See: :func:`request_until`.
        """
        return self.request_until('head', *args, **kwargs)

    def encode_json(self, data, opts):
        """
        Returns the given data encoded as JSON

        Args:
            data (any): The data to be encoded.
            opts (dict): A dictionary populated with request context details.

        Returns:
            str
        """
        return json.dumps(data, indent=4)

    def encode_yaml(self, data, opts):
        """
        Returns the given data encoded as YAML

        Args:
            data (any): The data to be encoded.
            opts (dict): A dictionary populated with request context details.

        Returns:
            str
        """
        return yaml.safe_dump(data, default_flow_style=False)


def _check_response_for_deprecation(path, response):
    """
    Inspects the given response's HTTP headers for a value that indicates the given path
    is deprecated.  If so, an exception will be raised.

    Args:
        path (str): The original path that was requested.

        response (:class:`requests.Response`): The HTTP response object to be inspected.

    Returns:
        None

    Raises:
        :class:`~performline.clients.rest.exceptions.DeprecatedEndpoint`
    """

    if response.headers.get('X-PerformLine-Deprecated') == '1':
        _notAfter = response.headers.get('X-PerformLine-Deprecated-After')
        notAfter = None
        afterMsg = ''

        if _notAfter is not None:
            notAfter = datetime.strptime(_notAfter, '%Y-%m-%dT%H:%M:%S.%f')
            afterMsg = ' It will stop working after {0}. '.format(notAfter)

        raise DeprecatedEndpoint('The API endpoint "{0}" has been marked deprecated.{1}'
                                 .format(path, afterMsg),
                                 cutoff=notAfter)



# Copyright (c) 2016, PerformLine, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the company nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL PERFORMLINE, INC. BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from __future__ import absolute_import
from .client import Client
from .embedded.stdlib.clients.rest.exceptions import ErrorResponse
from .cliutils import errout
from .products.common.cli.brands import brands
from .products.common.cli.campaigns import campaigns
from .products.common.cli.rules import rules
from .products.common.cli.trafficsources import sources
from .products.web.cli.pages import pages
from .products.callcenter.cli.calls import calls
from .products.chatscout.cli.chats import chats
import click


class State(object):
    def __init__(self, context, options={}):
        self.context = context
        self.token = options.get('token')
        self.url = options.get('url')
        self.prefix = options.get('prefix')
        self.loglevel = options.get('loglevel')
        self.fmt = options.get('format')

        if self.loglevel is not None:
            self.client = Client(self.token, loglevel=self.loglevel)
        else:
            self.client = Client(self.token)

        if self.url:
            self.client.url = self.url

        if self.prefix:
            self.client.prefix = self.prefix


@click.group()
@click.option('--api-key', '-k',
              metavar='KEY',
              help=(
                'The API key to use for authentication '
                '[$PERFORMLINE_API_KEY]'
              ),
              envvar='API_KEY')
@click.option('--api-host',
              metavar='URL',
              help=(
                  'The URL to use to connect to the PerformLine API service '
                  '(this should normally not need to be changed)\n'
                  '[$PERFORMLINE_API_URL]'
              ),
              envvar='API_URL')
@click.option('--api-endpoint-prefix',
              metavar='PATH',
              help=(
                  'A string to prefix API endpoint paths with '
                  '(this should normally not need to be changed)\n'
                  '[$PERFORMLINE_API_PREFIX]'
              ),
              envvar='API_PREFIX')
@click.option('--log-level', '-L',
              default='warning',
              help=(
                  'Set the level of logging verbosity\n'
                  '[$PERFORMLINE_LOGLEVEL]'
              ),
              type=click.Choice([
                'debug', 'info', 'warning', 'error',
              ]),
              envvar='LOGLEVEL')
@click.option('--format', '-f',
              default='yaml',
              type=click.Choice([
                'yaml', 'json',
              ]))
@click.pass_context
def main(ctx,
         api_key,
         api_host,
         api_endpoint_prefix,
         log_level,
         format):
    """PerformLine Compliance API client utility"""
    ctx.obj = State(ctx, {
        'token': api_key,
        'url': api_host,
        'prefix': api_endpoint_prefix,
        'loglevel': log_level,
        'format': format,
    })


main.add_command(brands)
main.add_command(calls)
main.add_command(campaigns)
main.add_command(chats)
main.add_command(pages)
main.add_command(rules)
main.add_command(sources)

try:
    main(auto_envvar_prefix='PERFORMLINE')
except Exception as e:
    if isinstance(e, ErrorResponse):
        errout('The PerformLine API encountered an error: %s' % e.message)
    else:
        raise



from __future__ import absolute_import
import math
import json
from ...utils.dicts import deep_get, must_deep_get


class Response(object):
    """
    A base class that provides common accessors for all standard REST responses (successful or
    erroneous).  Normally, this class is not created directly, but is instead returned (usually
    subclassed) by :func:`~performline.clients.rest.StandardRestClient.request`.

    Args:
        response (:class:`requests.Response`): The response object from the underlying HTTP
            library.

        data (dict): The decoded response payload matching the standard REST response format.
    """

    def __init__(self, response, data={}):
        self.response = response
        self._payload = data

    @property
    def payload(self):
        """
        Attempts to retrieve an arbitrary payload from the embedded response data.  If the embedded
        data is a string, attemp to JSON decode it, but fallback to just returning the data as-is.

        If the cached value exists and is not none, return it.

        In all other cases, return None.

        Returns:
            The embedded payload as a decoded JSON value (list, dict), the data value as-is,
            or `None`.
        """
        if hasattr(self, '_payload') and self._payload is not None:
            return self._payload
        elif hasattr(self, 'data'):
            if isinstance(self.data, str):
                try:
                    self._payload = json.loads(self.data)
                except ValueError:
                    self._payload = self.data
            else:
                self._payload = self.data

            return self._payload

        return None

    def clear_payload(self):
        """
        Clears the internal cached payload object
        """
        self._payload = None

    @property
    def status_code(self):
        """
        Returns the HTTP status code embedded in the response payload.

        Returns:
            int
        """

        if self.response:
            fallback = self.response.status_code
        else:
            fallback = None

        return self.deep_get('StatusCode', fallback)

    @property
    def response_status(self):
        """
        Returns the response status as described in the standard REST 'Status' field.

        Returns:
            str; "success" or "error"
        """
        return self.must_deep_get('Status')

    @property
    def metadata(self):
        """
        Returns any metadata included in the response object.

        Returns:
            dict
        """
        return self.deep_get('Metadata', {})

    def metadata_get(self, key, fallback=None):
        """
        Returns a deeply nested value from the metadata object, or a fallback value if not present.

        Args:
            key (str): The path to the value to retrieve.

            fallback (any, optional): A value to return if required is False and the key isn't
                present.

        Returns:
            any
        """
        return deep_get(self.metadata, key, fallback)

    def haserror(self):
        """
        Shorthand for checking if the status is erroneous.

        Returns:
            bool
        """
        return (self.deep_get('Status') == 'error')

    def must_deep_get(self, path):
        """
        See: :func:`performline.utils.dicts.must_deep_get`
        """
        return must_deep_get(self.payload, path)

    def deep_get(self, path, fallback=None):
        """
        See: :func:`performline.utils.dicts.deep_get`
        """
        return deep_get(self.payload, path, fallback)


class SuccessResponse(Response):
    """
    A subclass of :class:`Response` that provides accessors for fields present in successful REST
    responses.
    """
    @property
    def total_length(self):
        """
        Return the total number of results in a multi-valued result set.

        Returns:
            int
        """
        return self.deep_get('ResultCount/Total', 0)

    @property
    def length(self):
        """
        Return the number of results in the current response, which may be less than
             :func:`total_length` in a paginated result set.

        Returns:
            int
        """
        return self.deep_get('ResultCount/Current', self.total_length)

    @property
    def limit(self):
        """
        Return the number of results per page.

        Returns:
            int
        """
        return self.deep_get('ResultCount/Limit', self.length)

    @property
    def offset(self):
        """
        Return the number of results the results list is offset from the beginning of
        the total resultset.

        Returns:
            int
        """
        return self.deep_get('ResultCount/Offset', 0)

    @property
    def total_pages(self):
        """
        Return the total number of pages that contain the full result set.

        Returns:
            int
        """
        if self.limit == 0:
            return 0
        else:
            return int(math.ceil(float(self.total_length) / float(self.limit)))

    @property
    def current_page(self):
        """
        Return the current page number in a paginated result set (starting with
        page 1).

        Returns:
            int
        """

        if self.offset == 0:
            return 1
        else:
            return int(float(self.offset)/float(self.limit))+1

    def results(self, index=None):
        """
        Return results from a standard REST response.

        Returns:
            list, dict (if ``index`` was specified)
        """
        results = self.must_deep_get('Results')

        if index is not None:
            if index < len(results):
                return results[index]
            else:
                return {}
        else:
            return results

    def results_get(self, key, fallback=None, flatten=False, required=False):
        """
        Retrieves a given value from each element in the Results response.

        Args:
            key (str): The path to the value to retrieve.

            fallback (any, optional): A value to return if required is False and the key isn't
                present.

            required (bool, optional): Whether to raise an exception if the key isn't present in
                any list element.

            flatten (bool, optional): If True, return None for empty results, and return the first
                list element if the result list has a single element, otherwise return the list.

        Returns:
            list, None, any

        Raises:
            :class:`~performline.utils.dicts.MissingKeyException`
        """

        values = []

        for result in self.deep_get('Results', []):
            if isinstance(result, dict):
                if required:
                    values.append(must_deep_get(result, key))
                else:
                    values.append(deep_get(result, key, fallback))

        if flatten:
            if len(values) == 0:
                return fallback
            elif len(values) == 1:
                return values[0]

        return values



# Copyright (c) 2016, PerformLine, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# * Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.
# * Neither the name of the company nor the
# names of its contributors may be used to endorse or promote products
# derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL PERFORMLINE, INC. BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

from setuptools import setup, find_packages
from performline import client_version

setup(
    name='performline',
    description='Client library for using the public PerformLine Compliance API',
    version=client_version,
    author='PerformLine Engineering',
    author_email='tech+api@performline.com',
    url='https://github.com/PerformLine/python-performline-client',
    install_requires=[
        'click',
        'requests',
        'pyyaml',
        'six',
    ],
    entry_points={
        'console_scripts': [
            'performline=performline.cli:main',
        ],
    },
    packages=find_packages(exclude=['*.tests']),
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Natural Language :: English',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Office/Business',
        'Environment :: Console',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: POSIX',
        'Operating System :: POSIX :: Linux',
        'Operating System :: Microsoft :: Windows',
        'Intended Audience :: Customer Service',
        'Intended Audience :: Financial and Insurance Industry',
    ],
)



"""
String formatting and manipulation functions.
"""
from __future__ import absolute_import
import re


WORD_OMIT = r'[\W\s\_\-]+'
WORD_SPLIT = re.compile(r'(?:([A-Z][^A-Z]*)|'+WORD_OMIT+')')
CAMEL_STRIP = re.compile(WORD_OMIT)


def camelize(value, upperFirst=False):
    """
    Convert a string into "camelCase" or "PascalCase".

    Args:
        value (str): The string to convert.

        upperFirst (bool, optional): Whether to produce "camelCase" or
            "PascalCase" (first letter capitalized).

    Returns:
        str
    """

    # split on word-separating characters (and discard them), or on capital
    # letters (preserving them)
    value = WORD_SPLIT.split(value)

    # filter out None and empty/only-whitespace strings from the list
    value = [x for x in value if x is not None and not x.strip() == '']

    # capitalize each word
    value = ''.join(x.title() for x in value)

    # do a final pass removing omitted characters from the value
    value = re.sub(WORD_OMIT, '', value)

    if upperFirst:
        # return PascalCase
        return value
    else:
        # return camelCase
        return value[0].lower() + value[1:]


def underscore(value, joiner='_'):
    """
    Convert a string into "snake_case" (words separated by underscores).

    Args:
        value (str): The string to convert.

        joiner (str, optional): The string to use when joining individual words
            if underscore (_) is not desired.

    Returns:
        str
    """
    # split on word-separating characters (and discard them), or on capital
    # letters (preserving them)
    value = WORD_SPLIT.split(value)

    # filter out None and empty/only-whitespace strings from the list
    value = [x for x in value if x is not None and not x.strip() == '']

    # lowercase each word and join
    return joiner.join(x.strip().lower() for x in value)



from __future__ import absolute_import
import httpretty
import json
from unittest import TestCase
from .models import RestModel
from . import StandardRestClient
from .utils import make_response


class TestGetAllModel(RestModel):
    rest_root = '/get-all/'


class TestGetByIdModel(RestModel):
    rest_root = '/get-by-id/'


class RestModelTest(TestCase):
    def test_formatted_path(self):
        model = RestModel(None, {
            'Id': 4,
        }, rest_root='/test/')

        self.assertEqual(model.formatted_path(), '/test/4/')

        self.assertEqual(model.formatted_path({
            'Id': 5,
        }), '/test/4/')

        self.assertEqual(model.formatted_path({
            'id': 6,
        }), '/test/6/')

        model.primary_key = 'Id'

        self.assertEqual(model.formatted_path({
            'Id': 5,
        }), '/test/5/')

        with self.assertRaises(KeyError):
            self.assertEqual(model.formatted_path(), '/test/5/')

    def test_get_all(self):
        httpretty.enable()
        httpretty.register_uri(httpretty.GET,
                               "http://test-rest-getall/get-all/",
                               content_type='application/json',
                               body=json.dumps(make_response([{
                                   'Id': 1,
                               }, {
                                   'Id': 2,
                               }, {
                                   'Id': 3,
                               }])))

        httpretty.register_uri(httpretty.GET,
                               "http://test-rest-getall/get-all/1/",
                               content_type='application/json',
                               body=json.dumps(make_response({
                                   'Id': 1,
                                   'Name': 'First',
                               })))

        httpretty.register_uri(httpretty.GET,
                               "http://test-rest-getall/get-all/2/",
                               content_type='application/json',
                               body=json.dumps(make_response({
                                   'Id': 2,
                                   'Name': 'Second',
                               })))

        httpretty.register_uri(httpretty.GET,
                               "http://test-rest-getall/get-all/3/",
                               content_type='application/json',
                               body=json.dumps(make_response({
                                   'Id': 3,
                                   'Name': 'Third',
                               })))

        client = StandardRestClient(url='http://test-rest-getall')

        items = TestGetAllModel.all(client)

        httpretty.disable()

        self.assertEqual(len(items), 3)

        self.assertIsInstance(items[0], TestGetAllModel)
        self.assertEqual(items[0].id, 1)
        self.assertEqual(items[0].Id, 1)
        self.assertEqual(items[0].name, 'First')
        self.assertEqual(items[0].Name, 'First')

        self.assertIsInstance(items[1], TestGetAllModel)
        self.assertEqual(items[1].id, 2)
        self.assertEqual(items[1].Id, 2)
        self.assertEqual(items[1].name, 'Second')
        self.assertEqual(items[1].Name, 'Second')

        self.assertIsInstance(items[2], TestGetAllModel)
        self.assertEqual(items[2].id, 3)
        self.assertEqual(items[2].Id, 3)
        self.assertEqual(items[2].name, 'Third')
        self.assertEqual(items[2].Name, 'Third')

    def test_get_all_no_autoload(self):
        httpretty.enable()
        httpretty.register_uri(httpretty.GET,
                               "http://test-rest-getall/get-all/",
                               content_type='application/json',
                               body=json.dumps(make_response([{
                                   'Id': 1,
                               }, {
                                   'Id': 2,
                               }, {
                                   'Id': 3,
                               }])))

        client = StandardRestClient(url='http://test-rest-getall')

        items = TestGetAllModel.all(client, autoload=False)

        httpretty.disable()

        self.assertEqual(len(items), 3)

        self.assertIsInstance(items[0], TestGetAllModel)
        self.assertEqual(items[0].id, 1)
        self.assertEqual(items[0].Id, 1)
        self.assertIsNone(items[0].name)
        self.assertIsNone(items[0].Name)

        self.assertIsInstance(items[1], TestGetAllModel)
        self.assertEqual(items[1].id, 2)
        self.assertEqual(items[1].Id, 2)
        self.assertIsNone(items[1].name)
        self.assertIsNone(items[1].Name)

        self.assertIsInstance(items[2], TestGetAllModel)
        self.assertEqual(items[2].id, 3)
        self.assertEqual(items[2].Id, 3)
        self.assertIsNone(items[2].name)
        self.assertIsNone(items[2].Name)

    def test_get_by_id(self):
        httpretty.enable()

        httpretty.register_uri(httpretty.GET,
                               "http://test-rest-getbyid/get-by-id/12345/",
                               content_type='application/json',
                               body=json.dumps(make_response({
                                   'Id': 12345,
                                   'Name': 'Test Data',
                               })))

        client = StandardRestClient(url='http://test-rest-getbyid')

        item = TestGetByIdModel.get(client, 12345)

        httpretty.disable()

        self.assertIsInstance(item, TestGetByIdModel)
        self.assertEqual(item.id, 12345)
        self.assertEqual(item.Id, 12345)
        self.assertEqual(item.name, 'Test Data')
        self.assertEqual(item.Name, 'Test Data')
